"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the cleaning interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from os import sep as directory_separator
from os import walk

import PyFunceble


class Clean:
    """
    Provide the cleaning logic.

    .. note::
        By cleaning we mean the cleaning of the :code:`output` directory.

    :param list_to_test: The list of domains we are testing.
    :type list_to_test: list|None

    :param bool clean_all:
        Tell the subsystem if we need to clean all.
        Which include, of course, the output directory but also
        all other file(s) generated by our system.
    :param str file_path:
        The path to the file we tested.

        .. note::
            This is only relevant if you use the MariaDB/MySQL database.
    """

    def __init__(self, clean_all=False, file_path=None):
        # We clean the output directory.
        self.almost_everything(clean_all, file_path)

    @classmethod
    def file_to_delete(cls, all_files=False):
        """
        Return the list of file to delete.
        """

        # We initiate the directory we have to look for.
        directory = "{0}{1}".format(
            PyFunceble.OUTPUT_DIRECTORY, PyFunceble.OUTPUTS.parent_directory
        )

        if not directory.endswith(directory_separator):  # pragma: no cover
            # For safety, if it does not ends with the directory separator, we append it
            # to its end.
            directory += directory_separator

        # We initiate a variable which will save the list of file to delete.
        result = []

        for root, _, files in walk(directory):
            # We walk in the directory and get all files and sub-directories.

            for file in files:
                # If there is files in the current sub-directory, we loop
                # through the list of files.

                if file in [".gitignore", ".keep"]:
                    continue

                if (
                    not all_files and "logs" in root and ".log" in file
                ):  # pragma: no cover
                    continue

                # The file is not into our list of file we do not have to delete.

                if root.endswith(directory_separator):
                    # The root ends with the directory separator.

                    # We construct the path and append the full path to the result.
                    result.append(root + file)
                else:
                    # The root directory does not ends with the directory separator.

                    # We construct the path by appending the directory separator
                    # between the root and the filename and append the full path to
                    # the result.
                    result.append(root + directory_separator + file)  # pragma: no cover

        # We return our list of file to delete.
        return result

    @classmethod
    def databases_to_delete(cls):  # pragma: no cover
        """
        Set the databases files to delete.
        """

        # We initate the result variable.
        result = []

        if PyFunceble.CONFIGURATION.db_type == "json":
            # We initiate the directory we have to look for.
            directory = PyFunceble.CONFIG_DIRECTORY

            # We append the dir_structure file.
            result.append(
                "{0}{1}".format(
                    directory, PyFunceble.OUTPUTS.default_files.dir_structure
                )
            )

            # We append the iana file.
            result.append(
                "{0}{1}".format(directory, PyFunceble.OUTPUTS.default_files.iana)
            )

            # We append the public suffix file.
            result.append(
                "{0}{1}".format(
                    directory, PyFunceble.OUTPUTS.default_files.public_suffix
                )
            )

            # We append the inactive database file.
            result.append(
                "{0}{1}".format(directory, PyFunceble.OUTPUTS.default_files.inactive_db)
            )

            # We append the mining database file.
            result.append(
                "{0}{1}".format(directory, PyFunceble.OUTPUTS.default_files.mining)
            )

        return result

    def almost_everything(self, clean_all=False, file_path=False):
        """
        Delete almost all discovered files.

        :param bool clean_all:
            Tell the subsystem if we have to clean everything instesd
            of almost everything.
        """

        if (
            "do_not_clean" not in PyFunceble.INTERN
            or not PyFunceble.INTERN["do_not_clean"]
        ):
            # We get the list of file to delete.
            to_delete = self.file_to_delete(clean_all)

            if (
                not PyFunceble.abstracts.Version.is_local_cloned() and clean_all
            ):  # pragma: no cover
                to_delete.extend(self.databases_to_delete())

            for file in to_delete:
                # We loop through the list of file to delete.

                # And we delete the currently read file.
                PyFunceble.helpers.File(file).delete()

                PyFunceble.LOGGER.info(f"Deleted: {file}")

            if clean_all:  # pragma: no cover
                to_avoid = ["whois"]
            else:
                to_avoid = ["whois", "auto_continue", "inactive", "mining"]

            if not file_path:
                query = "DELETE FROM {0}"
            else:  # pragma: no cover
                query = "DELETE FROM {0} WHERE file_path = %(file_path)s"

            if PyFunceble.CONFIGURATION.db_type in [
                "mariadb",
                "mysql",
            ]:  # pragma: no cover

                with PyFunceble.engine.MySQL() as connection:
                    for database_name in [
                        y
                        for x, y in PyFunceble.engine.MySQL.tables.items()
                        if x not in to_avoid
                    ]:
                        lquery = query.format(database_name)

                        with connection.cursor() as cursor:
                            cursor.execute(lquery, {"file_path": file_path})

                            PyFunceble.LOGGER.info(
                                "Cleaned the data related to "
                                f"{repr(file_path)} from the {database_name} table."
                            )

            if (
                not PyFunceble.abstracts.Version.is_local_cloned() and clean_all
            ):  # pragma: no cover
                PyFunceble.load_config()

                PyFunceble.LOGGER.info("Reloaded configuration.")



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the base to all CI engines.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys

import PyFunceble


class CIBase:
    """
    Provides a base for all CI engines.
    """

    authorized = False

    # We set the regex to match in order to bypass the execution of
    # PyFunceble.
    regex_bypass = r"\[PyFunceble\sskip\]"

    def __init__(self):
        self.git_email = PyFunceble.helpers.EnvironmentVariable("GIT_EMAIL").get_value(
            default=None
        )

        self.git_name = PyFunceble.helpers.EnvironmentVariable("GIT_NAME").get_value(
            default=None
        )

    def authorization(self):
        """
        Provides the operation authorization.
        """

        raise NotImplementedError()

    def permissions(self):
        """
        Provides a permission fixer.
        """

    def git_var_checker(self):
        """
        Process some checks of the git related variables.
        """

        if self.authorized:
            if not self.git_email:
                raise PyFunceble.exceptions.GitEmailNotFound()

            if not self.git_name:
                raise PyFunceble.exceptions.GitNameNotFound()

    def init_git_remote_with_token(self, token):
        """
        Provides a simple way to initiate the git remote url.

        :param str token: A token with push access.
        """

        if self.authorized:
            remote = self.get_remote_destination()

            commands = [
                ("git remote rm origin", False),
                ("git remote add origin " f"https://{token}@{remote}", False),
                ("git remote update", False),
                ("git fetch", False),
            ]

            self.exec_commands(commands)

    def init_git(self, token):
        """
        Initiate the git project/repository.
        """

        if self.authorized:
            self.git_var_checker()
            self.init_git_remote_with_token(token)

            commands = [
                (f'git config --global user.email "{self.git_email}"', False),
                (f'git config --global user.name "{self.git_name}"', False),
                ("git config --global push.default simple", False),
                (f'git checkout "{PyFunceble.CONFIGURATION.ci_branch}"', False),
                (
                    f'git pull origin "{PyFunceble.CONFIGURATION.ci_distribution_branch}"',
                    False,
                ),
            ]

            self.exec_commands(commands)

    def init(self):
        """
        Init the CI machine/environment.
        """

        raise NotImplementedError()

    @classmethod
    def get_remote_destination(cls):
        """
        Provides the remote destination to use.
        """

        regex = r"(?:[a-z]+(?:\s+|\t+))(.*)(?:(?:\s+|\t+)\([a-z]+\))"
        remote_of_interest = [
            x
            for x in PyFunceble.helpers.Command("git remote -v").execute().splitlines()
            if "(fetch)" in x
        ][0]

        filtered = PyFunceble.helpers.Regex(regex).match(
            remote_of_interest, return_match=True, group=1
        )

        if filtered and "@" in filtered:
            return filtered[filtered.find("@") + 1 :]

        if filtered and "//" in filtered:
            return filtered[filtered.find("//") + 2 :]

        raise ValueError("Could not find remote.")

    @classmethod
    def exec_commands(cls, commands):
        """
        Execute a set of commands.

        :param list commands: A list of command to run.
        """

        for command, allow_stdout in commands:
            if allow_stdout or PyFunceble.LOGGER.authorized:
                PyFunceble.LOGGER.debug(f"Executing: {repr(command)}")
                PyFunceble.helpers.Command(command).run_to_stdout()
            else:
                PyFunceble.helpers.Command(command).execute()

    def bypass(self):
        """
        Stop everything if :code:`[PyFunceble skip]` is matched into
        the latest commit message.
        """

        if self.authorized and PyFunceble.helpers.Regex(self.regex_bypass).match(
            PyFunceble.helpers.Command("git log -1").execute(), return_match=False
        ):

            PyFunceble.LOGGER.info("Bypass given. Ending process.")

            self.end_commit()

    def push(self, exit_it=True):
        """
        Push.

        :param bool exit_it: Allow us to directly exit after pushing.
        """

        if self.authorized:

            commands = [
                (f"git push origin {PyFunceble.CONFIGURATION.ci_branch}", False)
            ]

            self.exec_commands(commands)

            if exit_it:
                sys.exit(0)

    def end_commit(self):
        """
        Commit and push at the very end of the test.
        """

        if self.authorized:
            PyFunceble.output.Percentage().log()
            self.permissions()

            commands = [
                ("git add --all", True),
                (
                    "git commit -a -m "
                    f'"{PyFunceble.CONFIGURATION.ci_autosave_final_commit} [ci skip]"',
                    True,
                ),
            ]

            if PyFunceble.CONFIGURATION.command_before_end:
                PyFunceble.LOGGER.info(
                    f"Executing: {PyFunceble.CONFIGURATION.command_before_end}"
                )

                PyFunceble.helpers.Command(
                    PyFunceble.CONFIGURATION.command_before_end
                ).run_to_stdout()
                self.permissions()

            self.exec_commands(commands)

            if (
                PyFunceble.CONFIGURATION.ci_distribution_branch
                != PyFunceble.CONFIGURATION.ci_branch
            ):
                self.push(exit_it=False)

                # We now merge to the destination branch.
                commands = [
                    (
                        f'git checkout "{PyFunceble.CONFIGURATION.ci_distribution_branch}"',
                        True,
                    ),
                    (f'git pull origin "{PyFunceble.CONFIGURATION.ci_branch}"', False),
                    (
                        f'git push origin "{PyFunceble.CONFIGURATION.ci_distribution_branch}"',
                        False,
                    ),
                ]

                self.exec_commands(commands)
                sys.exit(0)
            else:
                self.push()

    def not_end_commit(self):
        """
        Commit and push at on the middle of the test.
        """

        if self.authorized:
            PyFunceble.output.Percentage().log()
            self.permissions()

            command = 'git add --all && git commit -a -m "{0}"'.format(
                PyFunceble.CONFIGURATION.ci_autosave_commit
            )

            if PyFunceble.CONFIGURATION.command:
                PyFunceble.LOGGER.info(f"Executing: {PyFunceble.CONFIGURATION.command}")

                PyFunceble.helpers.Command(
                    PyFunceble.CONFIGURATION.command
                ).run_to_stdout()
                self.permissions()

            PyFunceble.LOGGER.info(f"Executing: {command}")

            PyFunceble.helpers.Command(command).run_to_stdout()

            self.push()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides a way to convert our internal URL.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from PyFunceble.abstracts import Version
from PyFunceble.exceptions import WrongParameterType

from .base import ConverterBase


class InternalUrl(ConverterBase):
    """
    Converter of the internal URLs.

    .. note::
        The internal URLs are actually the URL that has nothing to
        do with what we are going to test.

        They are only relevant for the software itself.

    :param str data_to_convert: The data to convert
    """

    def __init__(self, data_to_convert):
        if not isinstance(data_to_convert, str):
            raise WrongParameterType(
                f"<data_to_convert> should be {str}, {type(data_to_convert)} given."
            )

        super().__init__(data_to_convert)

        self.converted_data = self.to_right_url()

    def to_right_url(self):
        """
        Process the conversion to the right URL.
        """

        if Version.is_local_dev():
            return self.data_to_convert.replace("master", "dev")
        return self.data_to_convert.replace("dev", "master")



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the inactive database interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime, timedelta

import PyFunceble


class InactiveDB:  # pylint: disable=too-many-instance-attributes
    """
    Provides the inactive database logic and interface.

    :param str filename: The name of the file we are processing.
    """

    is_present_cache = {}
    database = {}
    authorized = False
    filename = None

    def __init__(self, filename, parent_process=False):
        self.one_day = timedelta(days=1)
        self.database_file = ""

        self.authorized = self.authorization()
        self.parent = parent_process

        PyFunceble.LOGGER.debug(f"Authorization: {self.authorized}")

        if self.authorized:
            self.days = timedelta(days=PyFunceble.CONFIGURATION.days_between_db_retest)
            self.days_between_clean = timedelta(
                days=PyFunceble.CONFIGURATION.days_between_inactive_db_clean
            )

            if PyFunceble.CONFIGURATION.db_type == "json":
                self.database_file = "{0}{1}".format(
                    PyFunceble.CONFIG_DIRECTORY,
                    PyFunceble.OUTPUTS.default_files.inactive_db,
                )

            self.filename = filename

            self.table_name = self.get_table_name()
            self.to_retest = self.get_to_retest()

            PyFunceble.LOGGER.debug(f"Table Name: {self.table_name}")
            PyFunceble.LOGGER.debug(f"DB (File): {self.database_file}")

            self.initiate()

    def __contains__(self, subject):
        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if subject not in self.is_present_cache:
                    self.is_present_cache[subject] = False
                    if self[subject]:
                        self.is_present_cache[subject] = True

                return self.is_present_cache[subject]

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT COUNT(*) "
                    "FROM {0} "
                    "WHERE subject = %(subject)s AND file_path = %(file)s"
                ).format(self.table_name)

                with PyFunceble.engine.MySQL() as connection, connection.cursor() as cursor:
                    cursor.execute(query, {"subject": subject, "file": self.filename})

                    fetched = cursor.fetchone()

                return fetched["COUNT(*)"]

        return False  # pragma: no cover

    def __getitem__(self, subject):
        if (
            self.authorized
            and PyFunceble.CONFIGURATION.db_type == "json"
            and self.filename in self.database
            and subject in self.database[self.filename]
        ):
            return self.database[self.filename][subject]
        return {}

    def __setitem__(self, subject, data):
        if PyFunceble.CONFIGURATION.db_type == "json":
            actual_state = self[subject]

            if actual_state:
                if isinstance(actual_state, dict):  # pragma: no cover
                    if isinstance(data, dict):
                        self.database[self.filename][
                            subject
                        ] = PyFunceble.helpers.Merge(data).into(
                            self.database[self.filename][subject]
                        )
                    else:  # pragma: no cover
                        self.database[self.filename][subject] = data
                elif isinstance(actual_state, list):  # pragma: no cover
                    if isinstance(data, list):
                        PyFunceble.helpers.Merge(data).into(
                            self.database[self.filename][subject], strict=False
                        )
                    else:  # pragma: no cover
                        self.database[self.filename][subject].append(data)

                    self.database[self.filename][subject] = PyFunceble.helpers.List(
                        self.database[self.filename][subject]
                    ).format()
                else:  # pragma: no cover
                    self.database[self.filename][subject] = data
            else:
                if self.filename not in self.database:
                    self.database[self.filename] = {subject: data}
                else:
                    self.database[self.filename][subject] = data

    def __delitem__(self, subject):
        if PyFunceble.CONFIGURATION.db_type == "json" and self[subject]:
            del self.database[self.filename][subject]

    @classmethod
    def authorization(cls):
        """
        Provides the execution authorization.
        """

        return PyFunceble.CONFIGURATION.inactive_database

    @classmethod
    def get_table_name(cls):
        """
        Returns the name of the table to use.
        """

        if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
            return PyFunceble.engine.MySQL.tables["inactive"]
        return "inactive"

    def _merge(self):
        """
        Merges the database with the older one which
        has already been set into the database.
        """

        if self.authorized and PyFunceble.CONFIGURATION.db_type == "json":
            database_content = PyFunceble.helpers.Dict().from_json_file(
                self.database_file
            )

            for database_top_key in [
                x for x in database_content.keys() if database_content[x]
            ]:

                for database_low_key, data in database_content[
                    database_top_key
                ].items():
                    to_set = {}

                    if database_low_key.isdigit():
                        last_test_date = datetime.fromtimestamp(float(database_low_key))

                        for subject, status in data.items():
                            to_set[subject] = {
                                "included_at_epoch": last_test_date.timestamp(),
                                "included_at_iso": last_test_date.isoformat(),
                                "last_retested_at_epoch": last_test_date.timestamp(),
                                "last_retested_at_iso": last_test_date.isoformat(),
                                "status": status,
                            }
                    else:
                        to_set[database_low_key] = data

                    if database_top_key not in self.database:
                        self.database[database_top_key] = to_set
                    else:
                        self.database[database_top_key].update(to_set)

            PyFunceble.LOGGER.info("Merged possible old to the new format")

    def load(self):
        """
        Loads the content of the database file.
        """

        if self.authorized and PyFunceble.CONFIGURATION.db_type == "json":
            if PyFunceble.helpers.File(self.database_file).exists():
                self._merge()
            else:
                self.database = {self.filename: {}}

            if self.filename not in self.database:  # pragma: no cover
                self.database[self.filename] = {}

            PyFunceble.LOGGER.info(
                "Database content loaded in memory. (DATASET WONT BE LOGGED)"
            )

    def save(self):
        """
        Saves the current database into the database file.
        """

        if (
            self.authorized
            and self.parent
            and PyFunceble.CONFIGURATION.db_type == "json"
        ):
            PyFunceble.helpers.Dict(self.database).to_json_file(self.database_file)

            PyFunceble.LOGGER.info(f"Saved database into {repr(self.database_file)}.")

    def clean(self):
        """
        Cleans everything which is not needed anymore.
        """

        if self.authorized and self.parent:
            PyFunceble.LOGGER.info("Started to clean old entry from the database.")
            for subject in self.get_to_clean():
                self.remove(subject)
            PyFunceble.LOGGER.info("Finished to clean old entry from the database.")

    def initiate(self):
        """
        Initiates the databsse.
        """

        if self.authorized:
            self.load()
            self.clean()
            self.save()

    @classmethod
    def datetime(cls):
        """
        Gets the timestamp where we are going to save our current list.

        :return: The timestamp to append with the currently tested element.
        :rtype: int|str
        """

        return datetime.now()

    def add(self, subject, status):
        """
        Adds the given subject into the database.

        :param str subject: The subject we are working with.
        :param str status: The status of the given subject.
        """

        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                current_datetime = self.datetime()

                if self[subject]:
                    self[subject] = {
                        "status": status,
                        "last_retested_at_iso": current_datetime.isoformat(),
                        "last_retested_at_epoch": current_datetime.timestamp(),
                    }
                else:
                    self[subject] = {
                        "status": status,
                        "included_at_iso": current_datetime.isoformat(),
                        "last_retested_at_iso": current_datetime.isoformat(),
                        "included_at_epoch": current_datetime.timestamp(),
                        "last_retested_at_epoch": current_datetime.timestamp(),
                    }

                PyFunceble.LOGGER.info(
                    f"Indexed {repr(subject)} with the status "
                    f"{repr(status)} into {repr(self.filename)} database's."
                )

                self.save()
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                digest = PyFunceble.helpers.Hash(algo="sha256").data(
                    bytes(self.filename + subject, "utf-8")
                )

                query = (
                    "INSERT INTO {0} "
                    "(file_path, subject, status, digest) "
                    "VALUES (%(file)s, %(subject)s, %(status)s, %(digest)s)"
                ).format(self.table_name)

                with PyFunceble.engine.MySQL() as connection, connection.cursor() as cursor:
                    playload = {
                        "file": self.filename,
                        "subject": subject,
                        "status": status,
                        "digest": digest,
                    }

                    try:
                        cursor.execute(query, playload)

                        PyFunceble.LOGGER.info(
                            f"Inserted into the database: \n {playload}"
                        )
                    except PyFunceble.engine.MySQL.errors:
                        query = (
                            "UPDATE {0} "
                            "SET subject = %(subject)s, status = %(status)s "
                            "WHERE digest = %(digest)s"
                        ).format(self.table_name)

                        cursor.execute(
                            query,
                            {"subject": subject, "status": status, "digest": digest},
                        )

                        PyFunceble.LOGGER.info(
                            "Data already indexed, updated the modified "
                            f"column of the row related to {repr(subject)}."
                        )

    def remove(self, subject):
        """
        Removes all occurrences of the given subject from the database.

        :param str subject: The subject we are working with.
        """

        if self.authorized:
            if PyFunceble.CONFIGURATION.db_type == "json":
                if self[subject]:
                    del self[subject]

                    PyFunceble.LOGGER.info(
                        "Cleaned the data related to " f"{repr(subject)}."
                    )
                    self.save()
            elif PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "DELETE FROM {0} "
                    "WHERE file_path = %(file)s "
                    "AND subject = %(subject)s"
                ).format(self.table_name)

                with PyFunceble.engine.MySQL() as connection, connection.cursor() as cursor:
                    cursor.execute(query, {"file": self.filename, "subject": subject})

                    PyFunceble.LOGGER.info(
                        "Cleaned the data related to "
                        f"{repr(subject)} and {repr(self.filename)} from "
                        f"the {repr(self.table_name)} table."
                    )

    def __execute_query(self, query):  # pragma: no cover
        """
        Executes the query to get the list to retest or already tested.
        """

        with PyFunceble.engine.MySQL() as connection, connection.cursor() as cursor:
            cursor.execute(
                query,
                {
                    "file": self.filename,
                    "days": self.days,
                    "days_between_clean": self.days_between_clean,
                },
            )
            fetched = cursor.fetchall()

            if fetched:
                return {x["subject"] for x in fetched}

        return set()

    def get_to_retest(self):
        """
        Returns a set of subject to restest.
        """

        PyFunceble.LOGGER.info(
            "Getting the list of subjects to retest (DATASET WONT BE LOGGED)"
        )

        if (
            self.authorized
            and PyFunceble.CONFIGURATION.days_between_db_retest >= 0
            and self.filename in self.database
        ):
            if PyFunceble.CONFIGURATION.db_type == "json":
                result = set()

                for subject, info in self.database[self.filename].items():
                    if (
                        "last_retested_at_epoch" in info
                        and info["last_retested_at_epoch"]
                    ):
                        if (
                            datetime.now()
                            > datetime.fromtimestamp(info["last_retested_at_epoch"])
                            + self.days
                        ):
                            result.add(subject)
                    else:
                        result.add(subject)

                return result

            if PyFunceble.CONFIGURATION.db_type in ["mariadb", "mysql"]:
                query = (
                    "SELECT * FROM {0} WHERE file_path = %(file)s "
                    "AND CAST(UNIX_TIMESTAMP() AS {1}) "
                    "> (CAST(UNIX_TIMESTAMP(modified) AS {1}) + CAST(%(days)s AS {1}))"
                ).format(self.table_name, PyFunceble.engine.MySQL.get_int_cast_type())

                return self.__execute_query(query)
        return set()

    def get_already_tested(self):
        """
        Returns a set of already tested subjects.
        """

        PyFunceble.LOGGER.info(
            "Getting the list of already tested (DATASET WONT BE LOGGED)"
        )

        if (
            self.authorized
            and PyFunceble.CONFIGURATION.days_between_db_retest >= 0
            and self.filename in self.database
        ):
            if PyFunceble.CONFIGURATION.db_type == "json":
                result = set()

                for subject, info in self.database[self.filename].items():
                    if (
                        "last_retested_at_epoch" in info
                        and info["last_retested_at_epoch"]
                    ):
                        if (
                            datetime.now()
                            < datetime.fromtimestamp(info["last_retested_at_epoch"])
                            + self.days
                        ):
                            result.add(subject)
                    else:
                        result.add(subject)
                return result

            if PyFunceble.CONFIGURATION.db_type in [
                "mariadb",
                "mysql",
            ]:  # pragma: no cover
                query = (
                    "SELECT * FROM {0} WHERE file_path= %(file)s "
                    "AND CAST(UNIX_TIMESTAMP() AS {1}) "
                    "< (CAST(UNIX_TIMESTAMP(modified) AS {1}) + CAST(%(days)s AS {1}))"
                ).format(self.table_name, PyFunceble.engine.MySQL.get_int_cast_type())

                return self.__execute_query(query)
        return set()

    def get_to_clean(self):
        """
        Returns a set of subject to clean from the database.
        """

        PyFunceble.LOGGER.info(
            "Getting the list of subject to clean (DATASET WONT BE LOGGED)"
        )

        if (
            self.authorized
            and PyFunceble.CONFIGURATION.days_between_inactive_db_clean >= 0
            and self.filename in self.database
        ):
            if PyFunceble.CONFIGURATION.db_type == "json":
                result = set()

                for subject, info in self.database[self.filename].items():
                    if "included_at_epoch" in info and info["included_at_epoch"]:
                        if (
                            datetime.now()
                            > datetime.fromtimestamp(info["included_at_epoch"])
                            + self.days_between_clean
                        ):
                            result.add(subject)
                    else:
                        result.add(subject)

                return result

            if PyFunceble.CONFIGURATION.db_type in [
                "mariadb",
                "mysql",
            ]:  # pragma: no cover

                query = (
                    "SELECT * FROM {0} WHERE file_path= %(file)s "
                    "AND CAST(UNIX_TIMESTAMP() AS {1}) "
                    "> (CAST(UNIX_TIMESTAMP(created) AS {1}) + CAST(%(days_between_clean)s AS {1}))"
                ).format(self.table_name, PyFunceble.engine.MySQL.get_int_cast_type())

                return self.__execute_query(query)
        return set()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the status interface for domains and IP availability check.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble

from ..gatherer_base import GathererBase
from .extra_rules import ExtraRules


class DomainAndIp(GathererBase):
    """
    Gather the availability of the given
    IP or Domain.
    """

    # pylint: disable=no-member

    def __init__(self, subject, filename=None, whois_db=None, inactive_db=None):
        super().__init__(
            subject, filename=filename, whois_db=whois_db, inactive_db=inactive_db
        )

        self.subject_type += "domain"
        self.__gather()

    def __gather_expiration_date(self):
        """
        Gather the expiration date.
        """

        if not self.status.whois_server:
            return None, None

        if self.whois_db:
            expiration_date_from_database = self.whois_db.get_expiration_date(
                self.subject
            )
        else:
            expiration_date_from_database = None

        if expiration_date_from_database:
            return expiration_date_from_database, "DATE EXTRACTED FROM WHOIS DATABASE"

        whois_record = PyFunceble.lookup.Whois(
            self.subject,
            self.status.whois_server,
            timeout=PyFunceble.CONFIGURATION.timeout,
        ).request()

        try:
            expiration_date = PyFunceble.extractor.ExpirationDate(
                whois_record
            ).get_extracted()

            if expiration_date and not PyFunceble.helpers.Regex(
                r"[0-9]{2}\-[a-z]{3}\-2[0-9]{3}"
            ).match(expiration_date, return_match=False):
                # The formatted expiration date does not match our unified format.

                # We log the problem.
                PyFunceble.output.Logs().expiration_date(self.subject, expiration_date)

                PyFunceble.LOGGER.error(
                    "Expiration date of "
                    f"{repr(self.subject)} ({repr(expiration_date)}) "
                    "was not converted proprely."
                )

            if self.whois_db:
                # We save the whois record into the database.
                self.whois_db.add(self.subject, expiration_date, whois_record)
        except PyFunceble.exceptions.WrongParameterType:
            expiration_date = None

        return expiration_date, whois_record

    def __gather_extra_rules(self):
        """
        Handle the lack of WHOIS record.
        """

        if self.status["_status"].lower() not in PyFunceble.STATUS.list.invalid:
            if self.status["_status"].lower() in PyFunceble.STATUS.list.down:
                self.status.dns_lookup = PyFunceble.DNSLOOKUP.request(self.subject)

                if self.status.dns_lookup:
                    self.status["_status"] = PyFunceble.STATUS.official.up
                    self.status["_status_source"] = "DNSLOOKUP"
                else:
                    self.status["_status"] = PyFunceble.STATUS.official.down
                    self.status["_status_source"] = "DNSLOOKUP"
        else:
            self.status.dns_lookup = PyFunceble.DNSLOOKUP.request(self.subject)

            if self.status.dns_lookup:
                # This is a safety. Indeed, as I may not be that reactive in the
                # future, this allow a new rule to be checked with the dns lookup
                # logic.

                self.status["_status"] = PyFunceble.STATUS.official.up
                self.status["_status_source"] = "DNSLOOKUP"

        PyFunceble.LOGGER.debug(
            f'[{self.subject}] State before extra rules:\n{self.status["_status"]}'
        )

        self.gather_http_status_code()

        self.status.status, self.status.status_source = ExtraRules(
            self.subject, self.subject_type, self.status.http_status_code
        ).handle(self.status["_status"], self.status["_status_source"])

    def __gather(self):
        """
        Process the gathering.
        """

        ip_validation = (
            self.status.ipv4_syntax_validation or self.status.ipv6_syntax_validation
        )

        if (
            PyFunceble.CONFIGURATION.local
            or self.status.domain_syntax_validation
            or ip_validation
        ):
            if not self.status.subdomain_syntax_validation:
                (
                    self.status.expiration_date,
                    self.status.whois_record,
                ) = self.__gather_expiration_date()

                if isinstance(self.status.expiration_date, str):
                    self.status["_status_source"] = "WHOIS"
                    self.status["_status"] = PyFunceble.STATUS.official.up
                    self.__gather_extra_rules()
                else:
                    self.status["_status_source"] = "WHOIS"
                    self.status["_status"] = PyFunceble.STATUS.official.down
                    self.__gather_extra_rules()
            else:
                self.status["_status_source"] = "SYNTAX"
                self.status["_status"] = PyFunceble.STATUS.official.down
                self.__gather_extra_rules()
        else:
            self.status["_status_source"] = "SYNTAX"
            self.status["_status"] = PyFunceble.STATUS.official.invalid

            self.__gather_extra_rules()

        PyFunceble.output.Generate(
            self.subject,
            self.subject_type,
            self.status.status,
            source=self.status.status_source,
            expiration_date=self.status.expiration_date,
            http_status_code=self.status.http_status_code,
            whois_server=self.status.whois_server,
            filename=self.filename,
            ip_validation=ip_validation,
        ).status_file(
            exclude_file_generation=(
                self.exclude_file_generation
                and self.status.status not in [PyFunceble.STATUS.official.up]
            )
        )

        PyFunceble.LOGGER.debug(f"[{self.subject}] State:\n{self.status.get()}")



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.engine.mining

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests

import PyFunceble
from PyFunceble.engine import Mining


class TestMining(TestCase):
    """
    Tests of PyFunceble.engine.mining.
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        PyFunceble.load_config(custom={"db_type": "json"})
        PyFunceble.CONFIGURATION.mining = True

        self.file_to_test = "this_file_is_a_ghost"

        self.storage_file = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.mining
        )

        self.mining = Mining(self.file_to_test, parent_process=True)

        self.file_to_test_instance = PyFunceble.helpers.File(self.file_to_test)
        self.storage_file_instance = PyFunceble.helpers.File(self.storage_file)

        self.file_to_test_instance.delete()
        self.storage_file_instance.delete()

        self.excepted_content = {
            self.file_to_test: {
                "myètherwället.com": ["www.google.com", "www.facebook.com"]
            }
        }

    def tearDown(self):
        """
        Setups everything needed after the tests.
        """

        self.file_to_test_instance.delete()
        self.storage_file_instance.delete()

    def test_load_file_not_exist(self):
        """
        Tests the method which let us load the storage file
        for the case that the storage file does not exists.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        self.mining.load()

        expected = {self.file_to_test_instance.path: {}}

        self.assertEqual(expected, self.mining.database)

    def test_load_file_exist(self):
        """
        Tests the method which let us load the storage file
        for the case that the storage file exists.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        PyFunceble.helpers.Dict(self.excepted_content.copy()).to_json_file(
            self.storage_file_instance.path
        )

        expected = self.excepted_content.copy()

        self.mining.load()

        self.assertEqual(expected, self.mining.database)

    def test_add(self):
        """
        Tests the method which let us add a new entry.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        expected = {
            self.file_to_test_instance.path: {
                "www.google.com": ["facebook.com", "www.facebook.com"]
            }
        }

        self.mining["www.google.com"] = ["facebook.com", "www.facebook.com"]

        self.assertEqual(expected, self.mining.database)

        self.mining["www.google.com"] = ["github.com"]

        expected[self.file_to_test_instance.path]["www.google.com"].append("github.com")

        self.assertEqual(expected, self.mining.database)

    def test_remove(self):
        """
        Tests the method which let us remove a new entry.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        expected = {
            self.file_to_test_instance.path: {
                "myètherwället.com": ["www.facebook.com", "facebook.com"],
                "example.org": ["facebook.com"],
            }
        }
        self.mining.database = {self.file_to_test_instance.path: {}}

        self.mining["myètherwället.com"] = ["www.facebook.com", "facebook.com"]
        self.mining["example.org"] = ["www.facebook.com", "facebook.com"]

        self.mining.remove("example.org", "www.facebook.com")
        self.assertEqual(expected, self.mining.database)

    def test_list_of_mined(self):
        """
        Tests the method which let us get the list of previously mined
        data.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        self.mining.database = self.excepted_content.copy()

        expected = [
            ("myètherwället.com", "www.google.com"),
            ("myètherwället.com", "www.facebook.com"),
        ]

        self.assertEqual(expected, self.mining.list_of_mined())

    def test_save(self):
        """
        Tests the method which let us save the mined database.
        """

        expected = False
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        self.mining.database = self.excepted_content.copy()
        self.mining.save()

        expected = True
        actual = self.storage_file_instance.exists()

        self.assertEqual(expected, actual)

        expected = self.excepted_content.copy()
        actual = PyFunceble.helpers.Dict().from_json_file(
            self.storage_file_instance.path
        )

        self.assertEqual(
            expected, actual,
        )


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the public suffix database generation tool.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import PyFunceble


class PublicSuffix:  # pragma: no cover pylint: disable=too-few-public-methods
    """
    Let us generate a new version of the public suffix database.
    """

    database = {}

    def __init__(self):
        # We initiate the destination of our database.
        self.destination = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.public_suffix
        )

    @classmethod
    def _data(cls):
        """
        Get the database from the public suffix repository.
        """

        # We initiate a variable which will save the link to the upstream public suffix file.
        public_suffix_url = (
            "https://raw.githubusercontent.com/publicsuffix/list/%s/public_suffix_list.dat"
            % "master"
        )

        # And we return the content of the previously declared link.
        return PyFunceble.helpers.Download(public_suffix_url).text()

    def _extensions(self, line):
        """
        Extract the extension from the given line.

        :param str line: The line from the official public suffix repository.
        """

        # We strip the parsed line.
        line = line.strip()

        if not line.startswith("//") and "." in line:
            # * The parsed line is not a commented line.
            # and
            # * There is a point in the parsed line.
            line = line.encode("idna").decode("utf-8")

            if line.startswith("*."):
                # The parsed line start with `*.`.

                # We remove the first two characters.
                line = line[2:]

            # We we split the points and we get the last element.
            # Explanation: The idea behind this action is to
            # always get the extension.
            extension = line.split(".")[-1]

            if extension in self.database:
                # The extension is alrady in our database.

                # We update the content of the 1st level TDL with
                # the content of the suffix.
                # In between, we format so that we ensure that there is no
                # duplicate in the database index content.
                self.database[extension] = PyFunceble.helpers.List(
                    self.database[extension] + [line]
                ).format()
            else:
                # The extension is not already in our database.

                # We append the currently formatted extension and the line content.
                self.database.update({extension: [line]})

    def update(self):
        """
        Update of the content of the :code:`public-suffix.json`.
        """

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We print a message for the user on screen.
            print(
                f"Update of {PyFunceble.OUTPUTS.default_files.public_suffix}", end=" "
            )

        # We loop through the line of the upstream file.
        list(map(self._extensions, self._data().split("\n")))

        # We save the content of our database in the final testination.
        PyFunceble.helpers.Dict(self.database).to_json_file(self.destination)

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We inform the user that everything goes right.
            print(PyFunceble.INTERN["done"])



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.abstracts.package

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests
from unittest.mock import patch

from PyFunceble.abstracts import Version


class TestVersion(TestCase):
    """
    Tests of PyFunceble.abstracts.Version
    """

    def test_split_version(self):
        """
        Tests the case that we want to split the version.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = ["1", "0", "0"]
        actual = Version.split_versions(given)

        self.assertEqual(expected, actual)

    def test_split_version_with_non_digits(self):
        """
        Tests the case that we want to split the version
        but also have the code name.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = (["1", "0", "0"], "dev (Hello, World!)")
        actual = Version.split_versions(given, return_non_digits=True)

        self.assertEqual(expected, actual)

    def test_literal_comparison(self):
        """
        Tests the literal comparison.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = True
        actual = Version.literally_compare(given, given)

        self.assertEqual(expected, actual)

    def test_literal_comparison_different(self):
        """
        Tests the litaral comparison for the case that both given version are different.
        """

        given = "1.0.0.dev (Hello, World!)"
        expected = False
        actual = Version.literally_compare(given, given.replace(".", "_"))

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "1.0.0.dev (Hello, World)")
    def test_compare_local_version_is_same(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "1.0.0.dev (Hello, World)"
        expected = None
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "1.50.0.dev (Hello, World)")
    def test_compare_local_version_is_older(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "2.34.0.dev (Hello, World)"
        expected = True
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0.dev (Hello, World)")
    def test_compare_local_version_is_newer(self):
        """
        Tests the comparison for the case that the local version is older.
        """

        given = "1.15.0.dev (Hello, World)"
        expected = False
        actual = Version.compare(given)

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0.dev (Hello, World)")
    def test_is_local_dev(self):
        """
        Tests if the local version is the dev one.
        """

        expected = True
        actual = Version.is_local_dev()

        self.assertEqual(expected, actual)

    @patch("PyFunceble.abstracts.Package.VERSION", "2.10.0. (Hello, World)")
    def test_is_not_local_dev(self):
        """
        Tests if the local version is the not the dev one.
        """

        expected = False
        actual = Version.is_local_dev()

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.database.inactive

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from datetime import datetime, timedelta
from unittest import TestCase
from unittest import main as launch_tests
from unittest.mock import Mock, patch

import PyFunceble
from PyFunceble.database.inactive import InactiveDB
from time_zone import TZ


class TestInactiveDB(TestCase):
    """
    Tests of PyFunceble.database.inactive
    """

    def setUp(self):
        """
        Setups everything needed for the tests.
        """

        PyFunceble.load_config(
            generate_directory_structure=False, custom={"db_type": "json"}
        )

        self.file_to_test = "this_is_a_ghost"
        self.storage_file = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.inactive_db
        )

        self.time_past = str(int((datetime.now() - timedelta(days=365)).timestamp()))
        self.time_future = str(int((datetime.now() + timedelta(days=365)).timestamp()))

        self.inactive_db = InactiveDB(self.file_to_test, parent_process=True)

        PyFunceble.helpers.File(self.storage_file).delete()

    def tearDown(self):
        """
        Setups everything needed after a test.
        """

        PyFunceble.helpers.File(self.storage_file).delete()

    def test_load_file_does_not_exists(self):
        """
        Tests the case that we load the file but it does not exists.
        """

        self.inactive_db.load()
        expected = {self.file_to_test: {}}

        self.assertEqual(expected, self.inactive_db.database)

    @patch("datetime.datetime")
    def test_load_file_exists(self, datetime_patch):
        """
        Tests the case that we load the file.
        """

        our_value = datetime(1970, 1, 1, 1, 3, 10, 0, tzinfo=TZ("+", hours=1).get())
        datetime_patch = Mock(wraps=datetime)
        datetime_patch.now = Mock(return_value=our_value)
        datetime_patch.fromtimestamp = Mock(return_value=our_value)
        patcher = patch("PyFunceble.database.inactive.datetime", new=datetime_patch)
        patcher.start()

        # We also test the merging.
        to_write = {
            self.file_to_test: {
                "190": {"example.com": PyFunceble.STATUS.official.invalid}
            },
            "this_is_a_well_informed_ghost": {
                "example.com": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        expected = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": our_value.timestamp(),
                    "included_at_iso": our_value.isoformat(),
                    "last_retested_at_epoch": our_value.timestamp(),
                    "last_retested_at_iso": our_value.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
            "this_is_a_well_informed_ghost": {
                "example.com": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        PyFunceble.helpers.Dict(to_write).to_json_file(self.storage_file)

        self.inactive_db.load()

        self.assertEqual(expected, self.inactive_db.database)

        patcher.stop()

    def test_clean(self):
        """
        Tests the cleaning process
        """

        to_write = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        PyFunceble.helpers.Dict(to_write).to_json_file(self.storage_file)

        self.inactive_db.load()
        self.inactive_db.parent = True
        self.inactive_db.clean()
        self.inactive_db.parent = False

        expected = {self.file_to_test: {}}

        self.assertEqual(expected, self.inactive_db.database)

    def test_get_to_retest(self):
        """
        Tests the method which is supposed to provides the list
        of subject to restest.
        """

        today = datetime.now()
        past = today - timedelta(days=100)
        to_write = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": today.timestamp(),
                    "included_at_iso": today.isoformat(),
                    "last_retested_at_epoch": today.timestamp(),
                    "last_retested_at_iso": today.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.org": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.net": {
                    "included_at_epoch": past.timestamp(),
                    "included_at_iso": past.isoformat(),
                    "last_retested_at_epoch": past.timestamp(),
                    "last_retested_at_iso": past.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        PyFunceble.helpers.Dict(to_write).to_json_file(self.storage_file)

        self.inactive_db.load()

        expected = {"example.org", "example.net"}

        self.assertEqual(expected, self.inactive_db.get_to_retest())

        self.inactive_db.remove("example.org")
        self.inactive_db.remove("example.net")

        expected = set()

        self.assertEqual(expected, self.inactive_db.get_to_retest())

    def test_get_already_tested(self):
        """
        Tests of the method which gives us the list of already
        tested subject.
        """

        today = datetime.now()
        past = today - timedelta(days=300)
        to_write = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": today.timestamp(),
                    "included_at_iso": today.isoformat(),
                    "last_retested_at_epoch": today.timestamp(),
                    "last_retested_at_iso": today.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.org": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.net": {
                    "included_at_epoch": past.timestamp(),
                    "included_at_iso": past.isoformat(),
                    "last_retested_at_epoch": past.timestamp(),
                    "last_retested_at_iso": past.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        PyFunceble.helpers.Dict(to_write).to_json_file(self.storage_file)

        self.inactive_db.load()

        expected = {"example.com", "example.org"}

        self.assertEqual(expected, self.inactive_db.get_already_tested())

        self.inactive_db.remove("example.com")
        self.inactive_db.remove("example.org")

        expected = set()

        self.assertEqual(expected, self.inactive_db.get_already_tested())

        self.inactive_db.authorized = False

        self.assertEqual(expected, self.inactive_db.get_already_tested())

    def test_get_to_clean(self):
        """
        Tests of the method which gives us the list of subject to clean.
        """

        today = datetime.now()
        past = today - timedelta(days=300)
        to_write = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": today.timestamp(),
                    "included_at_iso": today.isoformat(),
                    "last_retested_at_epoch": today.timestamp(),
                    "last_retested_at_iso": today.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.org": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "example.net": {
                    "included_at_epoch": past.timestamp(),
                    "included_at_iso": past.isoformat(),
                    "last_retested_at_epoch": past.timestamp(),
                    "last_retested_at_iso": past.isoformat(),
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        PyFunceble.helpers.Dict(to_write).to_json_file(self.storage_file)

        self.inactive_db.load()

        excepted = {"example.net", "example.org"}

        self.assertEqual(excepted, self.inactive_db.get_to_clean())

        excepted = set()
        self.inactive_db.authorized = False

        self.assertEqual(excepted, self.inactive_db.get_to_clean())

    def test_save(self):
        """
        Tests the case that we save the database.
        """

        expected = {
            self.file_to_test: {
                "example.com": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        self.inactive_db.database = expected.copy()
        self.inactive_db.save()

        self.assertEqual(
            expected, PyFunceble.helpers.Dict().from_json_file(self.storage_file)
        )

    def test_initiate_tested_path_does_not_exists(self):
        """
        Tests the initiate method for the case that the file
        we are testing is not indexed.
        """

        expected = {self.file_to_test: {}}

        self.inactive_db.database = {}
        self.inactive_db.initiate()

        self.assertEqual(expected, self.inactive_db.database)

    @patch("datetime.datetime")
    def test_add_tested_path_does_not_exists(self, datetime_patch):
        """
        Tests the addition method for the case that the flle
        we are testing is not indexed.
        """

        self.inactive_db.database = {}

        our_value = datetime(1970, 1, 1, 1, 0, 2, 0, tzinfo=TZ("+", hours=1).get())
        datetime_patch = Mock(wraps=datetime)
        datetime_patch.now = Mock(return_value=our_value)
        patcher = patch("PyFunceble.database.inactive.datetime", new=datetime_patch)
        patcher.start()

        subject = "hello.world"

        expected = {
            self.file_to_test: {
                subject: {
                    "included_at_epoch": our_value.timestamp(),
                    "included_at_iso": our_value.isoformat(),
                    "last_retested_at_epoch": our_value.timestamp(),
                    "last_retested_at_iso": our_value.isoformat(),
                    "status": PyFunceble.STATUS.official.down,
                }
            }
        }

        self.inactive_db.add(subject, PyFunceble.STATUS.official.down)
        self.assertEqual(expected, self.inactive_db.database)

        subject = "world.hello"

        expected[self.file_to_test][subject] = expected[self.file_to_test][
            "hello.world"
        ].copy()

        self.inactive_db.add(subject, PyFunceble.STATUS.official.down)
        self.assertEqual(expected, self.inactive_db.database)

        self.inactive_db.add(subject, PyFunceble.STATUS.official.down)
        self.assertEqual(expected, self.inactive_db.database)

        patcher.stop()

    def test_remove(self):
        """
        Tests of the deletion method.
        """

        subject = "hello.world"

        self.inactive_db.database = {
            self.file_to_test: {
                subject: {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        expected = {self.file_to_test: {}}

        self.inactive_db.remove(subject)
        self.assertEqual(expected, self.inactive_db.database)

    def test_is_present(self):
        """
        Tests the presence checker method.
        """

        subject = "hello.world"

        self.inactive_db.database = {
            self.file_to_test: {
                subject: {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.invalid,
                },
            },
        }

        expected = True
        actual = subject in self.inactive_db

        self.assertEqual(expected, actual)

        expected = False
        actual = "world.hello" in self.inactive_db

        self.assertEqual(expected, actual)

    def test_set_item_from_outside(self):
        """
        Tests the case that we adds items from outside the class.
        """

        expected = {
            self.file_to_test: {
                "hello.world": {
                    "included_at_epoch": 190.0,
                    "included_at_iso": "1970-01-01T01:03:10",
                    "last_retested_at_epoch": 190.0,
                    "last_retested_at_iso": "1970-01-01T01:03:10",
                    "status": PyFunceble.STATUS.official.invalid,
                },
                "world.hello": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.down,
                },
            },
        }

        self.inactive_db.database = {
            self.file_to_test: {
                "world.hello": {
                    "included_at_epoch": 0.0,
                    "included_at_iso": "1970-01-01T01:00:00",
                    "last_retested_at_epoch": 0.0,
                    "last_retested_at_iso": "1970-01-01T01:00:00",
                    "status": PyFunceble.STATUS.official.down,
                },
            },
        }

        self.inactive_db["hello.world"] = {
            "included_at_epoch": 190.0,
            "included_at_iso": "1970-01-01T01:03:10",
            "last_retested_at_epoch": 190.0,
            "last_retested_at_iso": "1970-01-01T01:03:10",
            "status": PyFunceble.STATUS.official.invalid,
        }

        self.assertEqual(expected, self.inactive_db.database)


if __name__ == "__main__":
    launch_tests()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.helpers.environement_variable

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from os import environ
from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.helpers import EnvironmentVariable


class TestEnvironmentVariable(TestCase):
    """
    Tests of PyFunceble.helpers.environement_variable
    """

    def test_name_exists(self):
        """
        Tests the case that a name exist.
        """

        environ["TEST"] = "test"

        expected = True
        actual = EnvironmentVariable("TEST").exists()

        self.assertEqual(expected, actual)

    def test_name_does_not_exists(self):
        """
        Tests the case that a name does not exist.
        """

        expected = False
        actual = EnvironmentVariable("HELLO,WORLD").exists()

        self.assertEqual(expected, actual)

    def test_get_value(self):
        """
        Tests the case that the value is needed.
        """

        environ["TEST"] = "Hello, World!"
        expected = "Hello, World!"
        actual = EnvironmentVariable("TEST").get_value()

        self.assertEqual(expected, actual)

    def test_get_default_value(self):
        """
        Tests the case that the environment variable does not
        exists and the value is needed.
        """

        expected = None
        actual = EnvironmentVariable("Hello,World").get_value()

        self.assertEqual(expected, actual)

        expected = True
        actual = EnvironmentVariable("Hello,World").get_value(default=True)

        self.assertEqual(expected, actual)

    def test_set_value(self):
        """
        Tests the case the we want to set the value of an environment variable.
        """

        expected_value = "Hello!"

        expected_output = True
        actual = EnvironmentVariable("TEST").set_value("Hello!")

        self.assertEqual(expected_output, actual)
        self.assertEqual(expected_value, environ["TEST"])

    def test_set_value_wrong_type(self):
        """
        Tests the case that we want to set a non string value.
        """

        self.assertRaises(TypeError, lambda: EnvironmentVariable("TEST").set_value(1))


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the custom logs interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime

import PyFunceble


class Logs:
    """
    Provide a clean and unique way to work with logs.
    Indeed, it's not good to have logs spread around the code :smile:

    :param str output: A path to the JSON file we are going to write.
    """

    def __init__(self, output=None):
        self.output = output
        self.current_time = str(datetime.now().timestamp())

    @classmethod
    def _get_content(cls, file):
        """
        Get and return the content of the given log file.

        :param str file: The file we have to get the content from.

        :return The content of the given file.
        :rtype: dict
        """

        if PyFunceble.helpers.File(file).exists():
            return PyFunceble.helpers.Dict().from_json_file(file)

        return {}

    @classmethod
    def _write_content(cls, content, file):
        """
        Write the content into the given file.

        :param str content: The dict to write.

        :param str file: The file to write.
        """

        if not PyFunceble.CONFIGURATION.no_files:
            if not isinstance(content, dict):
                content = {}

            PyFunceble.helpers.Dict(content).to_json_file(file)

    def whois(self, subject, record):
        """
        Logs the WHOIS record if needed.

        :param str subject: The currently tested subject.
        :param str record: The record to log.
        """

        if PyFunceble.CONFIGURATION.debug and PyFunceble.CONFIGURATION.logs:
            # The debug and the logs subsystem are activated.

            to_write = {self.current_time: {"domain": subject, "record": record}}

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.whois

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.debug(f"WHOIS Record of {repr(subject)}:\n{to_write}")

            self._write_content(current_content, output)

    def expiration_date(self, subject, extracted):
        """
        Logs the extracted expiration date.

        :param str subject: The currently tested subject.
        :param str extracted: The extracted expiration date (from WHOIS record).
        """

        if PyFunceble.CONFIGURATION.logs:
            # The logs subsystem is activated.

            if PyFunceble.INTERN["referer"]:
                referer = PyFunceble.INTERN["referer"]
            else:
                referer = None

            to_write = {
                self.current_time: {
                    "domain": subject,
                    "expiration_date": extracted,
                    "whois_server": referer,
                }
            }

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.date_format

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.critical(
                f"Wrong date format for {repr(subject)}:\n{to_write}"
            )

            self._write_content(current_content, output)

            if PyFunceble.CONFIGURATION.share_logs:
                # The logs sharing is activated.

                # And we share the logs with the api.
                PyFunceble.REQUESTS.post(
                    PyFunceble.LINKS.api_date_format,
                    data=to_write[self.current_time],
                    timeout=PyFunceble.CONFIGURATION.timeout,
                    verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                    allow_redirects=False,
                )

    def referer_not_found(self, subject, extension):
        """
        Logs the case that the referer was not found.

        :param str subject: The currently tested subject.
        :param str extension: The extension of the domain we are testing.
        """

        if PyFunceble.CONFIGURATION.logs:
            # The logs subsystem is activated.

            to_write = {self.current_time: {"domain": subject, "extension": extension}}

            if self.output:
                output = self.output
            else:
                output = PyFunceble.OUTPUT_DIRECTORY
                output += PyFunceble.OUTPUTS.parent_directory
                output += PyFunceble.OUTPUTS.logs.directories.parent
                output += PyFunceble.OUTPUTS.logs.filenames.no_referer

            current_content = self._get_content(output)
            current_content.update(to_write)

            PyFunceble.LOGGER.critical(
                f"Referer not found for {repr(subject)}:\n{to_write}"
            )

            self._write_content(current_content, output)

            if PyFunceble.CONFIGURATION.share_logs:
                # The logs sharing is activated.

                # And we share the logs with the api.
                PyFunceble.REQUESTS.post(
                    PyFunceble.LINKS.api_no_referer,
                    data=to_write[self.current_time],
                    timeout=PyFunceble.CONFIGURATION.timeout,
                    verify=PyFunceble.CONFIGURATION.verify_ssl_certificate,
                    allow_redirects=False,
                )



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the auto-save engine.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from datetime import datetime, timedelta

import PyFunceble

from .ci import GitLabCI, TravisCI


class AutoSave:  # pragma: no cover  pylint: disable=too-few-public-methods
    """
    Provides the autosave interface.
    """

    # We set the varible which will save the global authorization to operate.
    authorized = None
    # We set the variable which will save the starting time.
    start_time = None
    # We set the variable which will save the end time.
    endtime = None
    # We set the variable which will process as a chache for the
    # time exceeed value.
    time_exceed = False

    def __init__(self, start_time=None):
        self.current_ci_engine = self.get_current_ci()
        PyFunceble.LOGGER.debug(f"Current CI engine: {self.current_ci_engine}")

        if self.current_ci_engine:
            self.authorized = True

            PyFunceble.LOGGER.debug(
                "CI (already) Initiated: %s" % "ci_initiated" not in PyFunceble.INTERN
            )
            if "ci_initiated" not in PyFunceble.INTERN:
                self.current_ci_engine.init()
                self.current_ci_engine.bypass()

                if start_time is None:
                    self.start_time = datetime.now()
                else:
                    self.start_time = datetime.fromtimestamp(int(start_time))

                self.end_time = self.start_time + timedelta(
                    minutes=int(PyFunceble.CONFIGURATION.ci_autosave_minutes)
                )

                PyFunceble.INTERN["auto_save"] = {
                    "start_time": self.start_time.timestamp(),
                    "end_time": self.end_time.timestamp(),
                }

                PyFunceble.LOGGER.debug(f"Start Time: {self.start_time}")
                PyFunceble.LOGGER.debug(f"End Time:  {self.end_time}")

                PyFunceble.INTERN["ci_initiated"] = True
            elif "auto_save" in PyFunceble.INTERN:
                self.start_time = datetime.fromtimestamp(
                    PyFunceble.INTERN["auto_save"]["start_time"]
                )
                self.end_time = datetime.fromtimestamp(
                    PyFunceble.INTERN["auto_save"]["end_time"]
                )

                PyFunceble.LOGGER.debug(f"Start Time (Shared): {self.start_time}")
                PyFunceble.LOGGER.debug(f"End Time (Shared):  {self.end_time}")

    @classmethod
    def get_current_ci(cls):
        """
        Provides the current CI to use.
        """

        for ci_engine in [TravisCI(), GitLabCI()]:
            if ci_engine.authorized:
                return ci_engine

        return None

    def is_time_exceed(self):
        """
        Checks if the end time is exceed.
        """

        if self.authorized and hasattr(self, "end_time"):
            # We are authorized to operate.

            if not self.time_exceed and datetime.now() >= self.end_time:
                # * We did not tested previously if the time exceed.
                # and
                # * The time exceed.

                # We update the time exceed marker.
                self.time_exceed = True
        else:
            self.time_exceed = False

            PyFunceble.LOGGER.debug(f"Time exceed: {self.time_exceed}")

        return self.time_exceed

    def process(self, test_completed=False):
        """
        Processes the autosave base on the current state of the test.

        :param bool test_completed: Tell us if we finished the test.
        """

        if self.authorized:
            # We are authorized to operate.

            if test_completed:
                # The test was completed.

                # We run the end commit.
                self.current_ci_engine.end_commit()
            elif self.is_time_exceed():
                # The current time excessed the minimal time for autosaving.

                # We run the not end commit.
                self.current_ci_engine.not_end_commit()



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of the PyFunceble.helpers.regex

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.helpers import Regex


class TestRegex(TestCase):
    """
    Tests of the PyFunceble.helpers.regex.
    """

    def setUp(self):
        """
        Setups everything needed for the test.
        """

        self.data_list = [
            "hello",
            "world",
            "funilrys",
            "funceble",
            "PyFunceble",
            "pyfunceble",
        ]
        self.data = "Hello, this is Fun Ilrys. I just wanted to know how things goes around the tests."  # pylint: disable=line-too-long

    def test_not_matching_list(self):
        """
        Tests the method which let us get a list of non
        matching strin from a given list of string.
        """

        regex = "fun"
        expected = ["hello", "world", "PyFunceble"]
        actual = Regex(regex).get_not_matching_list(self.data_list)

        self.assertEqual(expected, actual)

    def test_matching_list(self):
        """
        Tests the method which let us get a list of
        matchint string from a given list of string.
        """

        regex = "fun"
        expected = ["funilrys", "funceble", "pyfunceble"]
        actual = Regex(regex).get_matching_list(self.data_list)

        self.assertEqual(expected, actual)

    def test_match_rematch(self):
        """
        Tests the matching method for the case that we want to rematch
        the different groups.
        """

        regex = r"([a-z]{1,})\s([a-z]{1,})\s"
        expected = "is"
        actual = Regex(regex).match(self.data, rematch=True, group=1)

        self.assertEqual(expected, actual)

    def test_match_get_group(self):
        """
        Tests the matching method for the case that we want
        a specific group.
        """

        regex = "e"
        expected = "e"
        actual = Regex(regex).match(self.data, group=0)

        self.assertEqual(expected, actual)

        regex = r"([a-z]{1,})\s([a-z]{1,})\s"
        expected = "this"
        actual = Regex(regex).match(self.data, group=1)

        self.assertEqual(expected, actual)

    def test_replace_no_replacement(self):
        """
        Tests the replacement method for the case that we replacement
        is not given.
        """

        regex = "th"
        expected = self.data
        actual = Regex(regex).replace_match(self.data, None)

        self.assertEqual(expected, actual)

    def test_replace(self):
        """
        Tests the replacement method.
        """

        regex = "th"
        expected = "Hello, htis is Fun Ilrys. I just wanted to know how htings goes around hte tests."  # pylint: disable=line-too-long
        actual = Regex(regex).replace_match(self.data, "ht")

        self.assertEqual(expected, actual)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the file helpers.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from os import path, remove, stat
from shutil import copy as shutil_copy
from shutil import move as shutil_move

from .directory import Directory


class File:
    """
    Simplify the file manipulations.

    :param str file_path: The file path to work with.
    """

    def __init__(self, file_path=None):
        self.path = file_path

    def exists(self, file_path=None):
        """
        Checks if the given file path exists.

        :param str file_path:
            The file path to check.

            .. note::
                If :code:`None` is given, we
                report to the globally given file path.
        :rtype: bool
        """
        if not file_path:
            file_path = self.path

        return path.isfile(file_path)

    def get_size(self, file_path=None):
        """
        Provides the size (in bytes) of the
        given file path.

        :param str file_path:
            The file path to check.

            .. note::
                If :code:`None` is given, we
                report to the globally given file path.
        :rtype: int
        """

        if not file_path:
            file_path = self.path

        return stat(file_path).st_size

    def is_empty(self, file_path=None):
        """
        Checks if the given file path is empty.
        """

        return self.get_size(file_path=file_path) <= 0

    def delete(self, file_path=None):
        """
        Deletes the given file path if it exists.

        :param str file_path:
            The file path to check.

            .. note::
                If :code:`None` is given, we
                report to the globally given file path.

        :return: The non existance state of the file.
        :rtype: bool
        """
        if not file_path:
            file_path = self.path

        if self.exists(file_path=file_path):
            remove(file_path)
        return not self.exists(file_path=file_path)

    def write(self, data, overwrite=False, encoding="utf-8", file_path=None):
        """
        Write the given data into the given file path.

        :param str data: The data to write.
        :param str encoding: The encoding to use while opening the file.
        :param str file_path:
            The file path to check.

            .. note::
                If :code:`None` is given, we
                report to the globally given file path.
        """

        if not file_path:
            file_path = self.path

        if isinstance(data, str):
            if overwrite or not self.exists(file_path=file_path):
                Directory(path.dirname(file_path)).create()

                with open(file_path, "w", encoding=encoding) as file_stream:
                    file_stream.write(data)
            else:
                with open(file_path, "a", encoding=encoding) as file_stream:
                    file_stream.write(data)

    def read(self, file_path=None, encoding="utf-8"):
        """
        Read the given file path and return it's content.

        :param str file_path:
            The file path to check.

            .. note::
                If :code:`None` is given, we
                report to the globally given file path.
        :param str encoding: The encoding to use.
        :rtype: str
        """

        if not file_path:
            file_path = self.path

        data = None

        if self.exists(file_path):
            with open(self.path, "r", encoding=encoding) as file_stream:
                data = file_stream.read()

        return data

    def copy(self, destination):
        """
        Copy the globaly given file path to the given destination.

        :param str destination: The destination of the copy.
        """

        if self.exists(self.path):
            shutil_copy(self.path, destination)

    def move(self, destination):  # pragma: no cover
        """
        Move the globally given file path to the given destination.

        :param str destination: The destination of the file.
        """

        if self.exists(self.path):
            shutil_move(self.path, destination)



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the status interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble


# pylint: disable=too-many-instance-attributes
class Status:
    """
    The status object.

    Provides the result of one of the status gatherer.

    :param str subject:
        The subject which we describe.

    :ivar str _status_source:
        The status source before the application
        of our extra rules (if activated).
    :ivar str _status:
        The status before the application of our
        extra rules (if activated).
    :ivar list dns_lookup:
        The result of the DNS Lookup logic.
    :ivar bool domain_syntax_validation:
        The domain syntax validation status.
    :ivar str expiration_date:
        The expiration date of the subject.
    :ivar str http_status_code:
        The HTTP status code of the subject.
    :ivar bool ipv4_range_syntax_validation:
        The IPv4 range validation status.
    :ivar bool ipv4_syntax_validation:
        The IPv4 syntax validation status.
    :ivar bool ipv6_range_syntax_validation:
        The IPv6 range validation status.
    :ivar bool ipv6_syntax_validation:
        The IPv6 syntax validation status.
    :ivar str status_source:
        The final status source.
    :ivar str status:
        The final status.
    :ivar bool subdomain_syntax_validation:
        The subdomain syntax validation status.
    :ivar str tested:
        The tested subject.
    :ivar bool url_syntax_validation:
        The URL syntax validation status.
    :ivar str whois_server:
        The WHOIS server we contact-ed to get
        the WHOIS record.
    :ivar str whois_record:
        The WHOIS record.
    """

    # pylint: disable=no-member, attribute-defined-outside-init

    resulting_indexes = [
        "_status_source",
        "_status",
        "dns_lookup",
        "domain_syntax_validation",
        "expiration_date",
        "http_status_code",
        "ipv4_range_syntax_validation",
        "ipv4_syntax_validation",
        "ipv6_range_syntax_validation",
        "ipv6_syntax_validation",
        "status_source",
        "status",
        "subdomain_syntax_validation",
        "tested",
        "url_syntax_validation",
        "whois_record",
        "whois_server",
    ]

    def __init__(self, subject):
        self.subject = subject
        self.checker = PyFunceble.Check(self.subject)

        pre_loading = {
            "_status_source": None,
            "_status": None,
            "dns_lookup": None,
            "domain_syntax_validation": self.checker.is_domain(),
            "expiration_date": None,
            "http_status_code": PyFunceble.HTTP_CODE.not_found_default,
            "ipv4_range_syntax_validation": self.checker.is_ipv4_range(),
            "ipv4_syntax_validation": self.checker.is_ipv4(),
            "ipv6_range_syntax_validation": self.checker.is_ipv6_range(),
            "ipv6_syntax_validation": self.checker.is_ipv6(),
            "status_source": None,
            "status": None,
            "subdomain_syntax_validation": self.checker.is_subdomain(),
            "tested": self.subject,
            "url_syntax_validation": self.checker.is_url(),
            "whois_record": None,
            "whois_server": PyFunceble.lookup.Referer(self.subject).get()[-1],
        }

        for description, value in pre_loading.items():
            setattr(self, description, value)

    def __getitem__(self, index):
        return getattr(self, index)

    def __setitem__(self, index, value):
        setattr(self, index, value)

    def get(self):
        """
        Provides the status in a dict format.
        """

        return {x: y for x, y in self.__dict__.items() if x in self.resulting_indexes}



# pylint:disable=line-too-long
"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Tests of PyFunceble.helpers.directory

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# pylint: enable=line-too-long

from os import getcwd, path
from os import sep as directory_separator
from unittest import TestCase
from unittest import main as launch_tests

from PyFunceble.helpers import Directory


class TestDirectory(TestCase):
    """
    Tests of PyFunceble.helpers.directory.
    """

    def test_get_current(self):
        """
        Tests the method which let us get the current directory.
        """

        expected = getcwd()
        actual = Directory.get_current()

        self.assertEqual(expected, actual)

    def test_get_current_with_end_sep(self):
        """
        Tests the method which let us get the current directory for the
        case that we want to have the directory separator at the end.
        """

        expected = getcwd() + directory_separator
        actual = Directory.get_current(with_end_sep=True)

        self.assertEqual(expected, actual)

    def tests_create_and_delete(self):
        """
        Tests the methods which let us create and delete
        a directory.
        """

        directory_test = "this_directory_is_a_ghost"
        dir_instance = Directory(directory_test)

        dir_instance.delete()

        expected = False
        actual = path.isdir(directory_test)

        self.assertEqual(expected, actual)

        dir_instance.create()

        expected = True
        actual = path.isdir(directory_test)

        self.assertEqual(expected, actual)

        dir_instance.delete()

    def test_fix_path(self):
        """
        Tests the method which allows us to fix
        directory paths.
        """

        expected = "hello" + directory_separator + "world" + directory_separator
        actual = Directory("/hello/world").fix_path()

        self.assertEqual(expected, actual)

        actual = Directory("\\hello\\world").fix_path()
        self.assertEqual(expected, actual)

        actual = Directory("hello\\world").fix_path()
        self.assertEqual(expected, actual)

        actual = Directory(r"hello\world").fix_path()
        self.assertEqual(expected, actual)

        actual = Directory(r"hello/world/").fix_path()
        self.assertEqual(expected, actual)

        to_test = ["", None, []]

        for element in to_test:
            actual = Directory(element).fix_path()
            self.assertEqual(element, actual)


if __name__ == "__main__":
    launch_tests()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides the IANA database generation tool.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


from multiprocessing import Pool

import PyFunceble


class Iana:  # pragma: no cover pylint: disable=too-few-public-methods
    """
    Logic behind the update and usage of `iana-domains-db.json`
    """

    # We map the list of server which have to be set manually because
    # they are not present into the IANA Root Zone Database.
    manual_server = {
        "aaa": "whois.nic.aaa",
        "abb": "whois.nic.abb",
        "able": "whois.nic.able",
        "accenture": "whois.nic.accenture",
        "aetna": "whois.nic.aetna",
        "aig": "whois.nic.aig",
        "americanexpress": "whois.nic.americanexpress",
        "amex": "whois.nic.amex",
        "amica": "whois.nic.amica",
        "amsterdam": "whois.nic.amsterdam",
        "analytics": "whois.nic.analytics",
        "aramco": "whois.nic.aramco",
        "arte": "whois.nic.arte",
        "as": "whois.nic.as",
        "athleta": "whois.nic.athleta",
        "audible": "whois.nic.audible",
        "author": "whois.nic.author",
        "aws": "whois.nic.aws",
        "axa": "whois.nic.axa",
        "azure": "whois.nic.azure",
        "baby": "whois.nic.baby",
        "banamex": "whois.nic.banamex",
        "bananarepublic": "whois.nic.bananarepublic",
        "baseball": "whois.nic.baseball",
        "bharti": "whois.nic.bharti",
        "bing": "whois.nic.bing",
        "bloomberg": "whois.nic.bloomberg",
        "bm": "whois.afilias-srs.net",
        "book": "whois.nic.book",
        "booking": "whois.nic.booking",
        "bot": "whois.nic.bot",
        "buzz": "whois.nic.buzz",
        "bz": "whois.afilias-grs.net",
        "call": "whois.nic.call",
        "calvinklein": "whois.nic.calvinklein",
        "caravan": "whois.nic.caravan",
        "cartier": "whois.nic.cartier",
        "caseih": "whois.nic.caseih",
        "cbn": "whois.nic.cbn",
        "cbre": "whois.nic.cbre",
        "cd": "chois.nic.cd",
        "chase": "whois.nic.chase",
        "circle": "whois.nic.circle",
        "cisco": "whois.nic.cisco",
        "citadel": "whois.nic.citadel",
        "citi": "whois.nic.citi",
        "citic": "whois.nic.citic",
        "cm": "whois.netcom.cm",
        "coupon": "whois.nic.coupon",
        "crown": "whois.nic.crown",
        "crs": "whois.nic.crs",
        "deal": "whois.nic.deal",
        "dealer": "whois.nic.dealer",
        "dell": "whois.nic.dell",
        "dhl": "whois.nic.dhl",
        "discover": "whois.nic.discover",
        "dnp": "whois.nic.dnp",
        "doosan": "whois.nic.doosan",
        "duns": "whois.nic.duns",
        "dupont": "whois.nic.dupont",
        "earth": "whois.nic.earth",
        "energy": "whois.nic.energy",
        "epost": "whois.nic.epost",
        "everbank": "whois.nic.everbank",
        "farmers": "whois.nic.farmers",
        "fast": "whois.nic.fast",
        "ferrero": "whois.nic.ferrero",
        "fire": "whois.nic.fire",
        "fj": "whois.usp.ac.fj",
        "flickr": "whois.nic.flickr",
        "flir": "whois.nic.flir",
        "food": "whois.nic.food",
        "ford": "whois.nic.ford",
        "fox": "whois.nic.fox",
        "free": "whois.nic.free",
        "frontier": "whois.nic.frontier",
        "ftr": "whois.nic.ftr",
        "ga": "whois.my.ga",
        "gap": "whois.nic.gap",
        "gh": "whois.nic.gh",
        "gmo": "whois.nic.gmo",
        "got": "whois.nic.got",
        "grainger": "whois.nic.grainger",
        "grocery": "whois.nic.grocery",
        "guardian": "whois.nic.guardian",
        "gucci": "whois.nic.gucci",
        "hair": "whois.nic.hair",
        "hbo": "whois.nic.hbo",
        "health": "whois.nic.health",
        "homegoods": "whois.nic.homegoods",
        "homesense": "whois.nic.homesense",
        "honeywell": "whois.nic.honeywell",
        "hot": "whois.nic.hot",
        "hoteles": "whois.nic.hoteles",
        "hotels": "whois.nic.hotels",
        "hotmail": "whois.nic.hotmail",
        "hsbc": "whois.nic.hsbc",
        "htc": "whois.nic.htc",
        "hyatt": "whois.nic.hyatt",
        "ieee": "whois.nic.ieee",
        "iinet": "whois.nic.iinet",
        "imdb": "whois.nic.imdb",
        "int": "whois.iana.org",
        "intel": "whois.nic.intel",
        "intuit": "whois.nic.intuit",
        "ipiranga": " whois.nic.ipiranga",
        "ipirange": "whois.nic.ipiranga",
        "itau": "whois.nic.itau",
        "iwc": "whois.nic.iwc",
        "jetzt": "whois.nic.jetzt",
        "jlc": "whois.nic.jlc",
        "jmp": "whois.nic.jmp",
        "jnj": "whois.nic.jnj",
        "jot": "whois.nic.jot",
        "joy": "whois.nic.joy",
        "jpmorgan": "whois.nic.jpmorgan",
        "jprs": "whois.nic.jprs",
        "kinder": "whois.nic.kinder",
        "kindle": "whois.nic.kindle",
        "kpmg": "whois.nic.kpmg",
        "kpn": "whois.nic.kpn",
        "kred": "whois.nic.kred",
        "kw": "whois.nic.kw",
        "lanxess": "whois.nic.lanxess",
        "lc": "whois2.afilias-grs.net",
        "lifeinsurance": "whois.nic.lifeinsurance",
        "like": "whois.nic.like",
        "lilly": "whois.nic.lilly",
        "lincoln": "whois.nic.lincoln",
        "living": "whois.nic.living",
        "lk": "whois.nic.lk",
        "loft": "whois.nic.loft",
        "lupin": "whois.nic.lupin",
        "maif": "whois.nic.maif",
        "marshalls": "whois.nic.marshalls",
        "mattel": "whois.nic.mattel",
        "mcd": "whois.nic.mcd",
        "mcdonalds": "whois.nic.mcdonalds",
        "merckmsd": "whois.nic.merckmsd",
        "microsoft": "whois.nic.microsoft",
        "mint": "whois.nic.mint",
        "mlb": "whois.nic.mlb",
        "mobily": "whois.nic.mobily",
        "moi": "whois.nic.moi",
        "montblanc": "whois.nic.montblanc",
        "moto": "whois.nic.moto",
        "msd": "whois.nic.msd",
        "mtpc": "whois.nic.mtpc",
        "mutual": "whois.nic.mutual",
        "mutuelle": "whois.nic.mutuelle",
        "nagoya": "whois.nic.nagoya",
        "nba": "whois.nic.nba",
        "netflix": "whois.nic.netflix",
        "neustar": "whois.nic.neustar",
        "nfl": "whois.nic.nfl",
        "nhk": "whois.nic.nhk",
        "nike": "whois.nic.nike",
        "northwesternmutual": "whois.nic.northwesternmutual",
        "now": " whois.nic.now",
        "ntt": "whois.nic.ntt",
        "nyc": "whois.nic.nyc",
        "office": "whois.nic.office",
        "okinawa": "whois.nic.okinawa",
        "oldnavy": "whois.nic.oldnavy",
        "open": "whois.nic.open",
        "orientexpress": "whois.nic.orientexpress",
        "otsuka": "whois.nic.otsuka",
        "passagens": "whois.nic.passagens",
        "pay": "whois.nic.pay",
        "pfizer": "whois.nic.pfizer",
        "pharmacy": " whois.nic.pharmacy",
        "piaget": " whois.nic.piaget",
        "pictet": "whois.nic.pictet",
        "pin": "whois.nic.pin",
        "ping": "whois.nic.ping",
        "pramerica": "whois.nic.pramerica",
        "praxi": "whois.nic.praxi",
        "prime": "whois.nic.prime",
        "pru": "whois.nic.pru",
        "prudential": "whois.nic.prudential",
        "ps": "whois.pnina.ps",
        "qvc": "whois.nic.qvc",
        "read": "whois.nic.read",
        "realtor": "whois.nic.realtor",
        "ren": "whois.nic.ren",
        "rocher": "whois.nic.rocher",
        "room": "whois.nic.room",
        "rw": "whois.ricta.org.rw",
        "ryukyu": "whois.nic.ryukyu",
        "safe": "whois.nic.safe",
        "safety": "whois.nic.safety",
        "sakura": "whois.nic.sakura",
        "sapo": "whois.nic.sapo",
        "sas": "whois.nic.sas",
        "save": "whois.nic.save",
        "secure": "whois.nic.secure",
        "sener": "whois.nic.sener",
        "shaw": "whois.afilias-srs.net",
        "shop": "whois.nic.shop",
        "silk": "whois.nic.silk",
        "skype": "whois.nic.skype",
        "sl": "whois.nic.sl",
        "smile": "whois.nic.smile",
        "sohu": "whois.nic.sohu",
        "song": "whois.nic.song",
        "spot": "whois.nic.spot",
        "staples": "whois.nic.staples",
        "statefarm": "whois.nic.statefarm",
        "stream": "whois.nic.stream",
        "suzuki": "whois.nic.suzuki",
        "swiftcover": "whois.nic.swiftcover",
        "talk": "whois.nic.talk",
        "taobao": "whois.nic.taobao",
        "target": "whois.nic.target",
        "tjmaxx": "whois.nic.tjmaxx",
        "tjx": "whois.nic.tjx",
        "tkmaxx": "whois.nic.tkmaxx",
        "tmall": "whois.nic.tmall",
        "tokyo": "whois.nic.tokyo",
        "tube": "whois.nic.tube",
        "tunes": "whois.nic.tunes",
        "tushu": "whois.nic.tushu",
        "tvs": "whois.nic.tvs",
        "unicom": "whois.nic.unicom",
        "uno": "whois.nic.uno",
        "vivo": "whois.nic.vivo",
        "vuelos": "whois.nic.vuelos",
        "wanggou": "whois.nic.wanggou",
        "watches": "whois.nic.watches",
        "weather": "whois.nic.weather",
        "weatherchannel": "whois.nic.weatherchannel",
        "weir": "whois.nic.weir",
        "whois": "whois.nic.qpon",
        "windows": "whois.nic.windows",
        "winners": "whois.nic.winners",
        "wow": "whois.nic.wow",
        "xbox": "whois.nic.xbox",
        "xn--1ck2e1b": "whois.nic.xn--1ck2e1b",
        "xn--2scrj9c": "whois.inregistry.net",
        "xn--3hcrj9c": "whois.inregistry.net",
        "xn--45br5cyl": "whois.inregistry.net",
        "xn--45brj9c": "whois.inregistry.net",
        "xn--8y0a063a": "whois.nic.xn--8y0a063a",
        "xn--bck1b9a5dre4c": "whois.nic.xn--bck1b9a5dre4c",
        "xn--cck2b3b": "whois.nic.xn--cck2b3b",
        "xn--czr694b": "whois.nic.xn--czr694b",
        "xn--e1a4c": "whois.eu",
        "xn--eckvdtc9d": "whois.nic.xn--eckvdtc9d",
        "xn--fct429k": "whois.nic.xn--fct429k",
        "xn--fpcrj9c3d": "whois.inregistry.net",
        "xn--fzc2c9e2c": "whois.nic.lk",
        "xn--g2xx48c": "whois.nic.xn--g2xx48c",
        "xn--gckr3f0f": "whois.nic.xn--gckr3f0f",
        "xn--gecrj9c": "whois.inregistry.net",
        "xn--gk3at1e": "whois.nic.xn--gk3at1e",
        "xn--h2breg3eve": "whois.inregistry.net",
        "xn--h2brj9c": "whois.inregistry.net",
        "xn--h2brj9c8c": "whois.inregistry.net",
        "xn--imr513n": "whois.nic.xn--imr513n",
        "xn--jvr189m": "whois.nic.xn--jvr189m",
        "xn--kpu716f": "whois.nic.xn--kpu716f",
        "xn--mgba3a3ejt": "whois.nic.xn--mgba3a3ejt",
        "xn--mgbb9fbpob": "whois.nic.xn--mgbb9fbpob",
        "xn--mgbbh1a": "whois.inregistry.net",
        "xn--mgbbh1a71e": "whois.inregistry.net",
        "xn--mgbgu82a": "whois.inregistry.net",
        "xn--nyqy26a": "whois.nic.xn--nyqy26a",
        "xn--otu796d": "whois.nic.xn--otu796d",
        "xn--pbt977c": "whois.nic.xn--pbt977c",
        "xn--rhqv96g": "whois.nic.xn--rhqv96g",
        "xn--rovu88b": "whois.nic.xn--rovu88b",
        "xn--rvc1e0am3e": "whois.inregistry.net",
        "xn--s9brj9c": "whois.inregistry.net",
        "xn--ses554g": "whois.registry.knet.cn",
        "xn--wgbh1c": "whois.dotmasr.eg",
        "xn--xkc2al3hye2a": "whois.nic.lk",
        "xn--xkc2dl3a5ee0h": "whois.inregistry.net",
        "yahoo": "whois.nic.yahoo",
        "yamaxun": "whois.nic.yamaxun",
        "yandex": "whois.nic.yandex",
        "yokohama": "whois.nic.yokohama",
        "you": "whois.nic.you",
        "za": "whois.registry.net.za",
        "zappos": "whois.nic.zappos",
        "zero": "whois.nic.zero",
        "zippo": "whois.nic.zippo",
    }

    def __init__(self):
        # We get the destination of the constructed IANA database.
        self.destination = (
            PyFunceble.CONFIG_DIRECTORY + PyFunceble.OUTPUTS.default_files.iana
        )

        if PyFunceble.helpers.File(self.destination).exists():
            # The destination exist.

            # We get its content.
            self.database = PyFunceble.helpers.Dict().from_json_file(self.destination)
        else:
            # The destination does not exist.

            # We initiate the local variable which will save the content of the database.
            self.database = {}

        # We initiate the URL to the IANA Root Zone Database page.
        self.iana_url = "https://www.iana.org/domains/root/db"

    def load(self):
        """
        Initiate the IANA database if it is not the case.
        """

        if "database" not in PyFunceble.INTERN or not PyFunceble.INTERN["database"]:
            # The global database is empty, None or does not exist.

            # We update it with the database content.
            PyFunceble.INTERN.update({"database": self.database.copy()})

    @classmethod
    def _get_referer(cls, extension):
        """
        Return the referer for the given extension.

        :param str extension: A valid domain extension.

        :return: The whois server to use to get the WHOIS record.
        :rtype: str
        """

        PyFunceble.LOGGER.info(f"Trying to find the referer of {repr(extension)}.")

        # We get the  whois record related to the domain extension we are currently
        # working with.
        iana_record = PyFunceble.lookup.Whois(
            "hello.{}".format(extension), PyFunceble.CONFIGURATION.iana_whois_server
        ).request()

        if iana_record and "refer" in iana_record:
            # The record is not empty.

            PyFunceble.LOGGER.info("Referer probably present. Trying to extract it.")

            # We initiate a regex which will extract the referer.
            regex_referer = r"(?s)refer\:\s+([a-zA-Z0-9._-]+)\n"

            # We try to extract the referer.
            matched = PyFunceble.helpers.Regex(regex_referer).match(
                iana_record, return_match=True, group=1
            )

            PyFunceble.LOGGER.debug(f"Extracted: {repr(matched)}")

            if matched:
                # The referer was extracted successfully.

                # We return the matched referer.
                return matched

        PyFunceble.LOGGER.info(
            "Could not extract nor find the referer from record. "
            "Trying to find it into our local entries."
        )

        # * The referer was not extracted successfully.
        # or
        # * The iana record is empty.

        if extension in cls.manual_server:
            # The extension is in the list of manual entries.

            PyFunceble.LOGGER.debug(
                f"Found local entry for {repr(extension)}: "
                f"{repr(cls.manual_server[extension])}"
            )

            # We return the server which we set manually.
            return cls.manual_server[extension]

        # We return None because we weren't able to get the server to call for
        # the given extension.
        return None

    @classmethod
    def _check_referer(cls, extension, referer):
        """
        Check if the given referer is reachable

        :param str extension: The extension the referer refers to.
        :param str referer: The referer to check.

        :rtype: bool
        """

        return (
            PyFunceble.lookup.Whois(
                "hello.{0}".format(extension), server=referer
            ).request()
            is not None
        )

    @classmethod
    def _get_extension_and_referer_from_block(cls, block):
        """
        Extract the extention from the given HTML block.
        Plus get its referer.

        :param str block: An HTML block.
        """

        # We extract the different extension from the currently readed line.
        regex_valid_extension = r"(/domains/root/db/)(.*)(\.html)"

        if "/domains/root/db/" in block:
            # The link is in the line.

            # We try to extract the extension.
            matched = PyFunceble.helpers.Regex(regex_valid_extension).match(
                block, return_match=True, rematch=True
            )[1]

            if matched:
                # The extraction is not empty or None.

                # We get the referer.
                referer = cls._get_referer(matched)

                if not referer:
                    referer = "whois.nic.{0}".format(matched)

                if cls._check_referer(matched, referer):
                    return matched, referer, True

                return matched, referer, False

        return None, None, None

    def update(self):
        """
        Update the content of `iana-domains-db.json` file.
        """

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We print on screen what we are doing.
            print(
                "Update of {0}".format(PyFunceble.OUTPUTS.default_files.iana), end=" "
            )

        upstream_lines = (
            PyFunceble.helpers.Download(self.iana_url)
            .text()
            .split('<span class="domain tld">')
        )

        with Pool(PyFunceble.CONFIGURATION.maximal_processes) as pool:
            already_checked = []
            for extension, referer, referer_checked in pool.map(
                self._get_extension_and_referer_from_block, upstream_lines
            ):
                if (extension is not None and referer) and (
                    referer_checked or referer in already_checked
                ):
                    if (
                        extension not in self.database
                        or self.database[extension] != referer
                    ):
                        # We add the extension to the databae.
                        self.database[extension] = referer

                already_checked.append(referer)

        # We save the content of the constructed database.
        PyFunceble.helpers.Dict(self.database).to_json_file(self.destination)

        if not PyFunceble.CONFIGURATION.quiet:
            # The quiet mode is not activated.

            # We indicate that the work is done without any issue.
            print(PyFunceble.INTERN["done"] + "\n")



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides an adblock decoding interface.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble.helpers as helpers
from PyFunceble.check import Check
from PyFunceble.exceptions import WrongParameterType

from .base import ConverterBase


class AdBlock(ConverterBase):
    """
    Converts an adblock filter line to a list os subject to test.
    """

    options_separator = "$"
    option_separator = ","

    def __init__(self, data_to_convert, aggressive=False):
        if not isinstance(data_to_convert, (str, list)):
            raise WrongParameterType(
                f"<data_to_convert> should be {str} or {list}, {type(data_to_convert)} given."
            )

        super().__init__(data_to_convert)
        self.aggressive = aggressive

    @classmethod
    def ignore_it(cls, subject):
        """
        Checks if we have to ignore the given subject.

        :param str subject: The subject ot work with.

        :return: The result of the check.
        :rtype: bool
        """

        # We set the list of regex to match to be
        # considered as ignored.
        #
        # Note: In a more aggressive way, r"(\$|,)(image)" may be added.
        to_ignore = [r"(^!|^@@|^\/|^\[|^\.|^-|^_|^\?|^&)"]

        for element in to_ignore:
            # We loop through the list of regex.

            if helpers.Regex(element).match(subject.strip(), return_match=False):
                # The currently read line match the currently read
                # regex.

                # We return true, it has to be ignored.
                return True

        # Wer return False, it does not has to be ignored.
        return False

    def remove_ignored(self, subject):
        """
        Removes the ignored element from the given list of subject.
        """

        if isinstance(subject, str):
            if self.ignore_it(subject):
                return []
            return [subject.strip()]

        return [x.strip() for x in subject if not self.ignore_it(x)]

    def extract_from_options(self, options):
        """
        Exctracts the relevant data from the list of options.

        :param list options: The list of options of a rule.

        :return: The list of domains.
        :rtype: list
        """

        result = []

        # We initiate the regex which will be used to extract the domain listed
        # under the option domain=
        regex_domain_option = r"domain=(.*)"

        for option in options:
            # We loop through the list of option.
            try:
                # We try to extract the list of domains from the currently read
                # option.
                domains = helpers.Regex(regex_domain_option).match(
                    option, return_match=True, rematch=True, group=0
                )[-1]

                if domains:
                    # We could extract something.

                    if self.aggressive:  # pragma: no cover
                        result.extend(
                            [
                                x
                                for x in domains.split("|")
                                if x and not x.startswith("~")
                            ]
                        )
                    else:
                        return True
            except TypeError:
                pass

        # We return the result.
        return result

    def extract_base(self, subject):
        """
        Extracts the base of the given element.

        As an example:
            given :code:`"hello.world/?is=beautiful"` returns :code:`"hello.world"`

        :param subject: The subject to work with.
        :type element: str|list
        """

        if isinstance(subject, list):
            return [self.extract_base(x) for x in subject]

        base = Check(subject).is_url(return_base=True)

        if base:
            return base

        if "/" in subject:
            return subject.split("/")[0]

        return subject

    def format_decoded(self, decoded, result=None):  # pragma: no cover
        """
        Formats the extracted adblock line in order to be
        compatible with what the system understand.

        :param str decoded: The decoded data to work with.

        :param list result: A list of the result of this method.

        :return: The list of domains or IP compatible with the system.
        :rtype: list
        """

        if result is None:
            result = []

        for data in decoded:
            if data:
                if "^" in data:
                    # There is an accent in the currently read line.

                    # We recall this method but with the current result state
                    # and splited data.
                    return self.format_decoded(data.split("^"), result)

                if "#" in data:
                    # There is a dash in the currently read line.

                    # We recall this method but with the current result state
                    # and splited data.
                    return self.format_decoded(data.split("#"), result)

                if "," in data:
                    # There is a comma in the currently read line.

                    # We recall this method but with the current result state
                    # and splited data.
                    return self.format_decoded(data.split(","), result)

                if "!" in data:
                    # There is an exclamation mark in the currently read line.

                    # We recall this method but with the current result state
                    # and splited data.
                    return self.format_decoded(data.split("!"), result)

                if "|" in data:
                    # There is a vertival bar in the currently read line.

                    # We recall this method but with the current result state
                    # and splited data.
                    return self.format_decoded(data.split("|"), result)

                if data:
                    # The currently read line is not empty.

                    data = self.extract_base(data)

                    # We create an instance of the checker.
                    checker = Check(data)

                    if data and (checker.is_domain() or checker.is_ip()):
                        # The extraced base is not empty.
                        # and
                        # * The currently read line is a valid domain.
                        # or
                        # * The currently read line is a valid IP.

                        # We append the currently read line to the result.
                        result.append(data)
                    elif data:
                        # * The currently read line is not a valid domain.
                        # or
                        # * The currently read line is not a valid IP.

                        # We try to get the url base.
                        url_base = checker.is_url(return_base=True)

                        if url_base:
                            # The url_base is not empty or equal to False or None.

                            # We append the url base to the result.
                            result.append(url_base)

        return result

    def __decode_v1(self, data):
        """
        Decodes the v1.

        :param str data: A string to decode.
        :rtype: list
        """

        result = []

        rematch = helpers.Regex(r"^(?:.*\|\|)([^\/\$\^]{1,}).*$").match(
            data, return_match=True, group=0, rematch=True
        )

        if rematch:
            if self.options_separator in data:
                options = data.split(self.options_separator)[-1].split(
                    self.option_separator
                )

                # pylint: disable=too-many-boolean-expressions
                if (
                    not options[-1]
                    or "third-party" in options
                    or "script" in options
                    or "popup" in options
                    or "xmlhttprequest" in options
                    or "all" in options
                    or "document" in options
                ):
                    result.extend(self.extract_base(rematch))

                extra = self.extract_from_options(options)

                if extra:
                    if isinstance(extra, list):  # pragma: no cover
                        extra.extend(self.extract_base(rematch))
                        result.extend(self.extract_base(extra))
                    else:
                        result.extend(self.extract_base(rematch))
            else:
                result.extend(self.extract_base(rematch))

        return result

    def __decode_v2(self, data):
        """
        Decodes the v2.

        :param str data: A string to decode.
        :rtype: list
        """

        result = []

        rematch = helpers.Regex(r"^\|(.*\..*)\|$").match(
            data, return_match=True, group=0, rematch=True
        )

        if rematch:
            result.extend(self.format_decoded(rematch))

        return result

    def __decode_v3(self, data):
        """
        Decodes the v3.

        :param str data: A string to decode.
        :rtype: list
        """

        result = []

        rematch = helpers.Regex(
            r"(?:#+(?:[a-z]+?)?\[[a-z]+(?:\^|\*)\=(?:\'|\"))(.*\..*)(?:(?:\'|\")\])"
        ).match(data, return_match=True, group=0, rematch=True)

        if rematch:
            result.extend(self.format_decoded(rematch))

        return result

    def __decode_v4(self, data):
        """
        Decodes the v4.

        :param str data: A string to decode.
        :rtype: list
        """

        result = []

        rematch = helpers.Regex(r"^(.*?)(?:#{2}|#@#)").match(
            data, return_match=True, group=0, rematch=True
        )

        if rematch:
            result.extend(self.format_decoded(rematch))

        return result

    def get_converted(self):
        """
        Converts and return the result of the conversion.

        :rtype: list
        """

        result = []

        for data in self.remove_ignored(self.data_to_convert):
            result.extend(self.__decode_v1(data))
            result.extend(self.__decode_v2(data))
            result.extend(self.__decode_v3(data))
            result.extend(self.__decode_v4(data))

        return helpers.List(result).format()



"""
The tool to check the availability or syntax of domains, IPv4, IPv6 or URL.

::


    ██████╗ ██╗   ██╗███████╗██╗   ██╗███╗   ██╗ ██████╗███████╗██████╗ ██╗     ███████╗
    ██╔══██╗╚██╗ ██╔╝██╔════╝██║   ██║████╗  ██║██╔════╝██╔════╝██╔══██╗██║     ██╔════╝
    ██████╔╝ ╚████╔╝ █████╗  ██║   ██║██╔██╗ ██║██║     █████╗  ██████╔╝██║     █████╗
    ██╔═══╝   ╚██╔╝  ██╔══╝  ██║   ██║██║╚██╗██║██║     ██╔══╝  ██╔══██╗██║     ██╔══╝
    ██║        ██║   ██║     ╚██████╔╝██║ ╚████║╚██████╗███████╗██████╔╝███████╗███████╗
    ╚═╝        ╚═╝   ╚═╝      ╚═════╝ ╚═╝  ╚═══╝ ╚═════╝╚══════╝╚═════╝ ╚══════╝╚══════╝

Provides our sorting engines.

Author:
    Nissar Chababy, @funilrys, contactTATAfunilrysTODTODcom

Special thanks:
    https://pyfunceble.github.io/special-thanks.html

Contributors:
    https://pyfunceble.github.io/contributors.html

Project link:
    https://github.com/funilrys/PyFunceble

Project documentation:
    https://pyfunceble.readthedocs.io/en/dev/

Project homepage:
    https://pyfunceble.github.io/

License:
::


    Copyright 2017, 2018, 2019, 2020 Nissar Chababy

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import PyFunceble


class Sort:  # pylint: disable=too-few-public-methods
    """
    Provides some sorting presets which we can
    parse to :py:class:`PyFunceble.helpers.list.List.custom_format`.
    """

    # We initiate a regex which will match everything which is not
    # a letter or a number.
    regex_replace = r"[^a-zA-Z0-9]"

    @classmethod
    def standard(cls, element):
        """
        Implements the standard and alphabetical sorting.

        :param str element: The element we are currently reading.

        :return: The formatted element.
        :rtype: str
        """

        # We remove all special characters and return the formatted string.
        return (
            PyFunceble.helpers.Regex(cls.regex_replace)
            .replace_match(element.strip(), "@funilrys")
            .replace("@funilrys", "")
        )

    @classmethod
    def hierarchical(cls, element):
        """
        The idea behind this method is to sort a list of domain hierarchicaly.

        :param str element: The element we are currently reading.

        :return: The formatted element.
        :rtype: str

        .. note::
            For a domain like :code:`aaa.bbb.ccc.tdl`.

            A normal sorting is done in the following order:
                1. :code:`aaa`
                2. :code:`bbb`
                3. :code:`ccc`
                4. :code:`tdl`

            This method allow the sorting to be done in the following order:
                1. :code:`tdl`
                2. :code:`ccc`
                3. :code:`bbb`
                4. :code:`aaa`

        """

        # We initiate a variable which will save the element to sort without
        # the extension.
        to_sort = ""

        # We initiate a variable which will save the full extension.
        full_extension = ""

        # We convert the parsed element to lower case.
        element = element.lower().strip()

        # We try to get the url base.
        url_base = PyFunceble.Check(element).is_url(return_base=True)

        if not isinstance(url_base, str):
            # The url base is not found.

            if "." in element:
                # There is point in the parsed element.

                # We get the position of the first letter of the extension.
                extension_index = element.rindex(".") + 1

                # We get the extension from the position of the first letter
                # of the extension.
                extension = element[extension_index:]

                if extension in PyFunceble.PSLOOOKUP:
                    # The extension is in the public suffix database.

                    for suffix in PyFunceble.PSLOOOKUP[extension]:
                        # We loop through the list of suffix of the extracted extension.

                        # We suffix the sufix with a point.
                        formatted_suffix = "." + suffix

                        if element.endswith(formatted_suffix):
                            # The elements ends with the suffix.

                            # We get the position of the first character of the suffix in
                            # the parsed element.
                            suffix_index = element.rindex(formatted_suffix)

                            # We update the to_sort variable with the element without the suffix.
                            to_sort = element[:suffix_index]

                            # We replace the full extension with the currently read suffix.
                            full_extension = suffix

                            # We break the loop, we got what we wanted.
                            break

                if not full_extension:
                    # The full extension is empty.

                    # We initiate it with the extension.
                    full_extension = element[extension_index:]

                    # We update the to_sort variable with the element without the extension.
                    to_sort = element[: extension_index - 1]

                # We append a point to the full extension because the point has to be
                # at the end and not at the begining of the extension.
                # To understand: Imagine a miror.
                full_extension += "."

                # We reverse the to_sort string.
                tros_ot = to_sort[::-1]

                if "." in tros_ot:
                    # There is a point in the reversed string.

                    # We prefix the full extension with the top level
                    # domain name.
                    full_extension = (
                        tros_ot[: tros_ot.index(".")][::-1] + "." + full_extension
                    )

                    # We remove the tor level domain from the rest of
                    # the reversed string.
                    tros_ot = tros_ot[tros_ot.index(".") + 1 :]

                    # * We reverse each level of the parsed element.
                    # and
                    # * We glue each level of the parsed element with each other.
                    #
                    # Note: after this, there is no point anymore.
                    reversion = full_extension + ".".join(
                        [x[::-1] for x in tros_ot.split(".")]
                    )

                    # We remove all special characters and return the formatted string.
                    return (
                        PyFunceble.helpers.Regex(cls.regex_replace)
                        .replace_match(reversion, "@funilrys")
                        .replace("@funilrys", "")
                    )

                # We remove all special characters and return the formatted string.
                return (
                    PyFunceble.helpers.Regex(cls.regex_replace)
                    .replace_match(to_sort + full_extension, "@funilrys")
                    .replace("@funilrys", "")
                )

            # There is no point in the parsed element.

            # We return the parsed element.
            return element

        # The url base is found.

        # We get the position of the element.
        protocol_position = element.rindex(url_base)

        # We extract the protocol from the element position.
        protocol = element[:protocol_position]

        # We return the output of this method but with the url base instead of the full url.
        return protocol + cls.hierarchical(url_base)
